/* $Id$ */
/**************************************************************************
 *
 *  FILE            netcamera.c
 *  MODULE OF       vidproxy
 *
 *  DESCRIPTION     Communication with camera server (c-grabber).
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <shhmsg.h>
#include <shhnet.h>

#include "shhsched.h"
#include "netviewer.h"
#include "netcamera.h"

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static unsigned char netcInPacket[MAX_PACKET];

static char netcLocalTcpPort[81] = "";
static int  netcLocalTcpInitial = -1;
static int  netcLocalTcp = -1;
static int  netcLocalUdp = -1;

static SchedId netcAcceptReadyId;
static SchedId netcTcpReadyId;
static SchedId netcUdpReadyId;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void
netcCloseConnection(void)
{
    if (netcLocalTcp >= 0) {
	schedRemoveDesc(netcTcpReadyId);
	netcTcpReadyId = 0;
	tcpClose(netcLocalTcp);
	netcLocalTcp = -1;
    }
    if (netcLocalUdp >= 0) {
	schedRemoveDesc(netcUdpReadyId);
	netcUdpReadyId = 0;
	udpClose(netcLocalUdp);
	netcLocalUdp = -1;
    }
}

static void
netcUdpInput(void *clientData, int fd)
{
    int n;

    udpRead(fd, netcInPacket, sizeof(netcInPacket), &n);
    netvSendPacket(netcInPacket, n);
}

static void
netcTcpInput(void *clientData, int fd)
{
    unsigned char buff[128];
    int           n, type;

    if (tcpRead(fd, buff, sizeof(buff), &n) != SN_OK || n == 0) {
	msgError("got error or empty packet, assuming broken connection\n");
	netcCloseConnection();
	return;
    }
    type = buff[0];
    switch (type) {
      case PACK_QUIT:
	msgMessage("shutting down connection on request\n");
        netcCloseConnection();
	break;
      default:
	msgError("unknown packet received (%u)\n", (unsigned) type);
    }
}

static void
netcAcceptConnection(void *clientData, int fd)
{
    int           sock;
    unsigned char buff[3];
    char          name[100];
    SNAddr        addr;
    SNPort        port;

    tcpAccept(fd, &sock);
    tcpLinger(sock, 1);
    if (netcLocalTcp >= 0) {
	buff[0] = PACK_HAVE_FEED;
	tcpWriteAll(sock, buff, 1);
	tcpClose(sock);
    } else {
	inGetPeerAddrPort(sock, &addr, &port);
	inHostAddrToName(addr, name, sizeof(name));
	msgMessage("camera daemon connected from %s\n", name);
	netcLocalTcp = sock;
	netcTcpReadyId
	    = schedAddDesc(SCHED_READ_READY, sock, netcTcpInput, NULL);
	udpClientSock(&netcLocalUdp);
	netcUdpReadyId
	    = schedAddDesc(SCHED_READ_READY, netcLocalUdp, netcUdpInput, NULL);
	inGetSockAddrPort(netcLocalUdp, &addr, &port);
	buff[0] = PACK_WELCOME_PORT;
	buff[1] = (port >> 8) & 0xFF;
	buff[2] = port & 0xFF;
	tcpWriteAll(sock, buff, 3);
	if (netvGetNumConnections() > 0)
	    netcStartGrab();
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void
netcSetLocalPort(const char *port)
{
    strcpy(netcLocalTcpPort, port);
}

void
netcInit(void)
{
    if (*netcLocalTcpPort == '\0')
	sprintf(netcLocalTcpPort, "%d", DEFAULT_CAMERA_PORT);

    msgMessage("setting up tcp-server for camera daemon at port %s\n",
	       netcLocalTcpPort);
    tcpServerOpen(netcLocalTcpPort, &netcLocalTcpInitial);
    netcAcceptReadyId
	= schedAddDesc(SCHED_READ_READY, netcLocalTcpInitial,
		       netcAcceptConnection, NULL);
    snSetErrorHandling(SN_REPORT_ERROR_AND_RETURN, NULL);
}

void
netcFinish(void)
{
    /* TODO: netcStopGrab(); */
    netcCloseConnection();
    if (netcLocalTcpInitial >= 0) {
	schedRemoveDesc(netcAcceptReadyId);
	tcpClose(netcLocalTcpInitial);
	netcLocalTcpInitial = -1;
    }
}

void
netcStartGrab(void)
{
    unsigned char buff[1];

    if (netcLocalTcp < 0)
	return;
    msgMessage("telling grabber to start grabbing\n");
    buff[0] = PACK_START_GRAB;
    tcpWriteAll(netcLocalTcp, buff, 1);
}

void
netcStopGrab(void)
{
    unsigned char buff[1];

    if (netcLocalTcp < 0)
	return;
    msgMessage("telling grabber to stop grabbing\n");
    buff[0] = PACK_STOP_GRAB;
    tcpWriteAll(netcLocalTcp, buff, 1);
}

int
netcCameraAvailable(void)
{
    return (netcLocalTcp >= 0);
}
