/* $Id$ */
/**************************************************************************
 *
 *  FILE            netviewer.c
 *  MODULE OF       vidproxy
 *
 *  DESCRIPTION     Communication with viewer clients (c-client and
 *                  SHHVid.java)
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <shhmsg.h>
#include <shhnet.h>

#include "shhsched.h"
#include "netcamera.h"
#include "netviewer.h"

#undef COUNT_PACKETS

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

#define MAX_CONNECTIONS 50

typedef struct {
    int     tcpSock;
    SchedId tcpSchedId;
    int     udpSock;
    SNAddr  udpAddr;
    SNPort  udpPort;
#ifdef COUNT_PACKETS
    unsigned long packetsSent;
#endif
} Connection;

static Connection conn[MAX_CONNECTIONS];
static int numConn = 0;

static char netvLocalTcpPort[81] = "";
static int  netvLocalTcpInitial = -1;

static SchedId netvAcceptReadyId;



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void netvTcpInput(void *clientData, int fd);

static int
netvAddConnection(int tcpSock, int udpSock)
{
    Connection *c;

    if (numConn == MAX_CONNECTIONS) {
	msgError("out of space for more connections\n");
	return -1;
    }
    c = &conn[numConn];
#ifdef COUNT_PACKETS
    c->packetsSent = 0;
#endif
    c->tcpSock = tcpSock;
    c->udpSock = udpSock;
    c->tcpSchedId
	= schedAddDesc(SCHED_READ_READY, tcpSock, netvTcpInput, NULL);
    if (++numConn == 1) {
	msgMessage("first connection, setting up grabber\n");
	netcStartGrab();
    }
    return numConn - 1;
}

static void
netvRemoveConnection(int tcpSock)
{
    int        q;
    Connection *c;

    for (q = 0; q < numConn; q++)
	if (conn[q].tcpSock == tcpSock)
	    break;
    if (q == numConn)
	return;
    c = &conn[q];
    schedRemoveDesc(c->tcpSchedId);
    if (c->tcpSock >= 0)
	tcpClose(c->tcpSock);
    if (c->udpSock >= 0)
	udpClose(c->udpSock);
    memcpy(c, &conn[numConn - 1], sizeof(Connection));
    if (--numConn == 0) {
	msgMessage("no connections left, stopping grabber\n");
	netcStopGrab();
    }
}

static int
netvFindConnection(int tcpSock)
{
    int q;

    for (q = 0; q < numConn; q++)
	if (conn[q].tcpSock == tcpSock)
	    return q;
    return -1;
}

static void
netvTcpInput(void *clientData, int fd)
{
    unsigned char buff[128];
    char          name[100];
    int           n, type, idx;
    SNPort        port;

    if (tcpRead(fd, buff, sizeof(buff), &n) != SN_OK || n == 0) {
	msgError("got error or empty packet, assuming broken connection\n");
	netvRemoveConnection(fd);
	return;
    }
    idx = netvFindConnection(fd);
    type = buff[0];
    switch (type) {
      case PACK_PORT:
	inGetPeerStr(fd, name, sizeof(name), 1);
	port = ((int) buff[1] << 8) | buff[2];
	conn[idx].udpPort = port;
	udpClientSock(&(conn[idx].udpSock));
	msgMessage("incoming call from %s using UDP port %u\n",
		   name, (unsigned) port);
	buff[0] = PACK_WELCOME;
	tcpWriteAll(fd, buff, 1);
	break;
      case PACK_HANGUP:
	msgMessage("shutting down connection on request\n");
        netvRemoveConnection(fd);
	break;
      default:
	msgError("unknown packet received (%u)\n", (unsigned) type);
    }
}

static void
netvAcceptConnection(void *clientData, int fd)
{
    int           sock, idx;
    unsigned char buff[1];
    SNAddr        addr;
    SNPort        port;

    tcpAccept(fd, &sock);
    tcpLinger(sock, 1);
    inGetPeerAddrPort(sock, &addr, &port);
    if (!netcCameraAvailable()) {
	buff[0] = PACK_NO_CAM;
	tcpWriteAll(sock, buff, 1);
    } else if ((idx = netvAddConnection(sock, -1)) < 0) {
	buff[0] = PACK_BUSY;
	tcpWriteAll(sock, buff, 1);
    } else {
	conn[idx].udpAddr = addr;
	conn[idx].tcpSchedId 
	    = schedAddDesc(SCHED_READ_READY, sock, netvTcpInput, NULL);
	buff[0] = PACK_SEND_PORT;
	tcpWriteAll(sock, buff, 1);
    }
}



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

void
netvSetLocalPort(const char *port)
{
    strcpy(netvLocalTcpPort, port);
}

void
netvInit(void)
{
    if (*netvLocalTcpPort == '\0')
	sprintf(netvLocalTcpPort, "%d", DEFAULT_VIEWER_PORT);

    msgMessage("setting up tcp-server for viewers at port %s\n",
	       netvLocalTcpPort);
    tcpServerOpen(netvLocalTcpPort, &netvLocalTcpInitial);
    netvAcceptReadyId
	= schedAddDesc(SCHED_READ_READY, netvLocalTcpInitial,
		       netvAcceptConnection, NULL);
    snSetErrorHandling(SN_REPORT_ERROR_AND_RETURN, NULL);
}

void
netvFinish(void)
{
    int q;

    for (q = 0; q < numConn; q++) {
	if (conn[q].tcpSock >= 0)
	    tcpClose(conn[q].tcpSock);
	if (conn[q].udpSock >= 0)
	    tcpClose(conn[q].tcpSock);
	schedRemoveDesc(conn[q].tcpSchedId);
    }
    numConn = 0;
    if (netvLocalTcpInitial >= 0) {
	tcpClose(netvLocalTcpInitial);
	netvLocalTcpInitial = -1;
	schedRemoveDesc(netvAcceptReadyId);
    }
}

void
netvSendPacket(const unsigned char *packet, int len)
{
    int        q, n;
    Connection *c;
#ifdef COUNT_PACKETS
    static int delay = 0;
    char name[81];
#endif

    if (!len)
	return;
    for (q = 0; q < numConn; q++) {
	c = &conn[q];
	if (c->udpSock >= 0) {
	    udpWriteTo(c->udpSock, packet, len, c->udpAddr, c->udpPort, &n);
	    if (n < len)
		msgError("partial packet sent\n");
	}
#ifdef COUNT_PACKETS
	++(c->packetsSent);
#endif
    }
#ifdef COUNT_PACKETS
    if (++delay == 100) {
	delay = 0;
	for (q = 0; q < numConn; q++) {
	    c = &conn[q];
	    inHostAddrToName(c->udpAddr, name, sizeof(name));
	    printf("sent %ld packets to %s\n", c->packetsSent, name);
	}
    }
#endif
}

int
netvGetNumConnections(void)
{
    return numConn;
}
