/* $Id$ */
/**************************************************************************
 *
 *  FILE            shhsched.c
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby <sverrehu@ifi.uio.no>
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#if defined(sgi)
#include <bstring.h>
#endif

#include <shhmsg.h>

#include "shhsched.h"

/* the following should be changed to a dynamic scheme later */
#define MAX_SCHED_TIME_FUNCS 20
#define MAX_SCHED_DESC_FUNCS FD_SETSIZE

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

typedef struct _SchedTimeItem {
    SchedId                id;
    SchedTimeFunc          func;
    void                  *clientData;
    struct timeval         when;
    struct _SchedTimeItem *next;
} SchedTimeItem;

typedef struct _SchedDescItem {
    SchedId                id;
    SchedDescFunc          func;
    void                  *clientData;
    int                    fd;
    SchedDescAction        action;
    struct _SchedDescItem *next;
} SchedDescItem;

static SchedTimeItem  schedTimePool[MAX_SCHED_TIME_FUNCS];
static SchedTimeItem *schedTimeFree;
static SchedTimeItem *schedTimeNext;

static SchedDescItem  schedDescReadPool[MAX_SCHED_DESC_FUNCS];
static SchedDescItem *schedDescReadFree;
static SchedDescItem *schedDescReadNext;
static SchedDescItem  schedDescWritePool[MAX_SCHED_DESC_FUNCS];
static SchedDescItem *schedDescWriteFree;
static SchedDescItem *schedDescWriteNext;
static SchedDescItem  schedDescExceptPool[MAX_SCHED_DESC_FUNCS];
static SchedDescItem *schedDescExceptFree;
static SchedDescItem *schedDescExceptNext;
static fd_set schedDescReadSet;
static fd_set schedDescWriteSet;
static fd_set schedDescExceptSet;

static SchedId schedLastId;

static int schedKeepLooping;



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

SchedId
schedAddTimeAt(struct timeval *when, SchedTimeFunc func, void *clientData)
{
    register SchedTimeItem *ti, *ti2;

    if ((ti = schedTimeFree) == NULL)
	msgFatal("too many functions scheduled for later calls\n");
    schedTimeFree = ti->next;
    ti->id = ++schedLastId;
    ti->func = func;
    ti->clientData = clientData;
    ti->when = *when;
    /* insert at correct location. assumes that tv_usec < 1 second */
    if ((ti2 = schedTimeNext) == NULL
	|| ti2->when.tv_sec > ti->when.tv_sec
	|| (ti2->when.tv_sec == ti->when.tv_sec
	    && ti2->when.tv_usec > ti->when.tv_usec)) {
	ti->next = schedTimeNext;
	schedTimeNext = ti;
    } else {
	while (ti2->next) {
	    if (ti2->next->when.tv_sec > ti->when.tv_sec
		|| (ti2->next->when.tv_sec == ti->when.tv_sec
		    && ti2->next->when.tv_usec > ti->when.tv_usec))
		break;
	    ti2 = ti2->next;
	}
	ti->next = ti2->next;
	ti2->next = ti;
    }
    return ti->id;
}

SchedId
schedAddTime(SchedTime delay, SchedTimeFunc func, void *clientData)
{
    struct timeval when;

    /* calculate the time to call this function */
    gettimeofday(&when, NULL);
    when.tv_sec += delay / 1000L;
    when.tv_usec += ((delay % 1000L) * 1000L);
    if (when.tv_usec >= 1000000L) {
	when.tv_usec -= 1000000L;
	++when.tv_sec;
    }
    return schedAddTimeAt(&when, func, clientData);
}

void
schedRemoveTime(SchedId id)
{
    SchedTimeItem *ti, **prev;

    if ((ti = schedTimeNext) == NULL)
	return;
    prev = &schedTimeNext;
    do {
	if (ti->id == id) {
	    *prev = ti->next;
	    ti->next = schedTimeFree;
	    schedTimeFree = ti;
	    break;
	}
	prev = &ti->next;
    } while ((ti = ti->next) != NULL);
}

void
schedRemoveTimeAll(void)
{
    int q;

    schedTimeNext = NULL;
    schedTimeFree = &schedTimePool[0];
    for (q = 0; q < MAX_SCHED_TIME_FUNCS - 1; q++)
	schedTimePool[q].next = &schedTimePool[q + 1];
    schedTimePool[q].next = NULL;
}

SchedId
schedAddDesc(SchedDescAction action, int fd,
	     SchedDescFunc func, void *clientData)
{
    SchedDescItem *di = NULL, **nextp = NULL, **freep = NULL;
    fd_set *fds = NULL;

    switch (action) {
      case SCHED_READ_READY:
	fds = &schedDescReadSet;
	freep = &schedDescReadFree;
	nextp = &schedDescReadNext;
	break;
      case SCHED_WRITE_READY:
	fds = &schedDescWriteSet;
	freep = &schedDescWriteFree;
	nextp = &schedDescWriteNext;
	break;
      case SCHED_EXCEPTION_READY:
	fds = &schedDescExceptSet;
	freep = &schedDescExceptFree;
	nextp = &schedDescExceptNext;
	break;
      default:
	msgFatal("illegal action in schedAddDesc\n");
    }
    if ((di = *freep) == NULL)
	msgFatal("too many descriptor functions scheduled\n");
    *freep = di->next;
    di->id = ++schedLastId;
    di->func = func;
    di->clientData = clientData;
    di->fd = fd;
    di->action = action;
    di->next = *nextp;
    *nextp = di;
    FD_SET(fd, fds);
    return di->id;
}

void
schedRemoveDesc(SchedId id)
{
    int q;
    SchedDescItem *di = NULL, **prev = NULL, **freep = NULL;
    fd_set *fds = NULL;

    for (q = 0; q < 3; q++) {
	switch (q) {
	  case 0:
	    di = schedDescReadNext;
	    prev = &schedDescReadNext;
	    freep = &schedDescReadFree;
	    fds = &schedDescReadSet;
	    break;
	  case 1:
	    di = schedDescWriteNext;
	    prev = &schedDescWriteNext;
	    freep = &schedDescWriteFree;
	    fds = &schedDescWriteSet;
	    break;
	  case 2:
	    di = schedDescExceptNext;
	    prev = &schedDescExceptNext;
	    freep = &schedDescExceptFree;
	    fds = &schedDescExceptSet;
	    break;
	}
	if (di == NULL)
	    continue;
	do {
	    if (di->id == id) {
		FD_CLR(di->fd, fds);
		*prev = di->next;
		di->next = *freep;
		*freep = di;
		goto finish;
	    }
	    prev = &di->next;
	} while ((di = di->next) != NULL);
    }
  finish:
}

void
schedRemoveDescAll(void)
{
    int q;

    schedDescReadNext = NULL;
    schedDescReadFree = &schedDescReadPool[0];
    for (q = 0; q < MAX_SCHED_DESC_FUNCS - 1; q++)
	schedDescReadPool[q].next = &schedDescReadPool[q + 1];
    schedDescReadPool[q].next = NULL;
    FD_ZERO(&schedDescReadSet);

    schedDescWriteNext = NULL;
    schedDescWriteFree = &schedDescWritePool[0];
    for (q = 0; q < MAX_SCHED_DESC_FUNCS - 1; q++)
	schedDescWritePool[q].next = &schedDescWritePool[q + 1];
    schedDescWritePool[q].next = NULL;
    FD_ZERO(&schedDescWriteSet);

    schedDescExceptNext = NULL;
    schedDescExceptFree = &schedDescExceptPool[0];
    for (q = 0; q < MAX_SCHED_DESC_FUNCS - 1; q++)
	schedDescExceptPool[q].next = &schedDescExceptPool[q + 1];
    schedDescExceptPool[q].next = NULL;
    FD_ZERO(&schedDescExceptSet);
}

void
schedInit(void)
{
    schedRemoveTimeAll();
    schedRemoveDescAll();
    schedLastId = 0;
}

void
schedFinish(void)
{
}

void
schedEndLoop(void)
{
    schedKeepLooping = 0;
}

/*-------------------------------------------------------------------------
 *
 *  NAME          schedLoop
 *
 *  FUNCTION      The main loop of the scheduler.
 *
 *  SYNOPSIS      #include "shhsched.h"
 *                void (void);
 *
 *  INPUT         
 *
 *  OUTPUT        
 *
 *  RETURNS       
 *
 *  DESCRIPTION   
 */
void 
schedLoop(void)
{
    int ret;
    /* the reason for using 0,1 instead of 0,0 for timeoutnow, is that
     * SGI select didn't return > 0 for 0,0. */
    struct timeval now, timeout, *tvp, timeoutnow = { 0, 0 };
    fd_set fdsr, fdsw, fdse;
    SchedTimeItem *ti;
    SchedDescItem *di;

    schedKeepLooping = 1;
    while (schedKeepLooping) {
	/* need a loop to restart the select if interrupted by a signal. */
	do {
	    /* find next time function to call. */
	    if ((ti = schedTimeNext) == NULL)
		tvp = NULL;  /* no function registered, sleep `forever' */
	    else {
		gettimeofday(&now, NULL);
		/* calculate timeout value */
		timeout.tv_sec = ti->when.tv_sec - now.tv_sec;
		if ((timeout.tv_usec = ti->when.tv_usec - now.tv_usec) < 0L) {
		    timeout.tv_usec += 1000000L;
		    --timeout.tv_sec;
		}
		/* check if already expired */
		if (timeout.tv_sec < 0 || timeout.tv_usec < 0)
		    tvp = &timeoutnow;
		else
		    tvp = &timeout;
	    }
	    fdsr = schedDescReadSet;
	    fdsw = schedDescWriteSet;
	    fdse = schedDescExceptSet;
	    ret = select(FD_SETSIZE, &fdsr, &fdsw, &fdse, tvp);
	    if (ret == 0) {
		/* timeout */
		/* call function */
		ti->func(ti->clientData);
		/* remove function from list */
		schedTimeNext = ti->next;
		ti->next = schedTimeFree;
		schedTimeFree = ti;
	    } else if (ret > 0) {
		/* some descriptor ready */
		/* first check the read set */
		di = schedDescReadNext;
		while (di) {
		    if (FD_ISSET(di->fd, &fdsr)) {
			di->func(di->clientData, di->fd);
			if (--ret == 0)
			    goto all_called;
		    }
		    di = di->next;
		}
		/* then check the write set */
		di = schedDescWriteNext;
		while (di) {
		    if (FD_ISSET(di->fd, &fdsw)) {
			di->func(di->clientData, di->fd);
			if (--ret == 0)
			    goto all_called;
		    }
		    di = di->next;
		}
		/* and finally check the exception set */
		di = schedDescExceptNext;
		while (di) {
		    if (FD_ISSET(di->fd, &fdse)) {
			di->func(di->clientData, di->fd);
			if (--ret == 0)
			    goto all_called;
		    }
		    di = di->next;
		}
	      all_called:
	    }
	} while (ret < 0 && errno == EINTR);
    }
}
